<?php

namespace Drupal\selective_better_exposed_filters\Plugin\better_exposed_filters\filter;

use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Plugin\views\filter\Bundle;
use Drupal\views\Plugin\views\filter\FilterPluginBase;
use Drupal\views\ViewExecutable;
use Drupal\views\Views;

/**
 * Base class for Better exposed filters widget plugins.
 */
abstract class SelectiveFilterBase {

  /**
   * {@inheritdoc}
   */
  public static function defaultConfiguration(): array {
    return [
      'options_show_only_used' => FALSE,
      'options_show_only_used_filtered' => FALSE,
      'options_hide_when_empty' => FALSE,
      'options_show_items_count' => FALSE,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function buildConfigurationForm(FilterPluginBase $filter, array $settings): array {
    if (!$filter->isExposed()) {
      return [];
    }

    $supported_plugins = [
      'taxonomy_index_tid',
      'bundle',
      'list_field',
    ];

    \Drupal::moduleHandler()->alter('selective_better_exposed_filters_supported_plugins', $supported_plugins);


    if (!in_array($filter->getPluginId(), $supported_plugins)) {
      return [];
    }

    return [
      'options_show_only_used' => [
        '#type' => 'checkbox',
        '#title' => t('Show only used items'),
        '#default_value' => !empty($settings['options_show_only_used']),
        '#description' => t('Restrict exposed filter values to those presented in the result set.'),
      ],
      'options_show_only_used_filtered' => [
        '#type' => 'checkbox',
        '#title' => t('Filter items based on filtered result set by other applied filters'),
        '#default_value' => !empty($settings['options_show_only_used_filtered']),
        '#description' => t('Restrict exposed filter values to those presented in the already filtered result set by other applied filters. (Not work when view use ajax form in block)'),
        '#states' => [
          'visible' => [
            ':input[name="exposed_form_options[bef][filter][' . $filter->field . '][configuration][options_show_only_used]"]' => [
              'checked' => TRUE,
            ],
          ],
        ],
      ],
      'options_hide_when_empty' => [
        '#type' => 'checkbox',
        '#title' => t('Hide filter, if no options'),
        '#default_value' => !empty($settings['options_hide_when_empty']),
        '#states' => [
          'visible' => [
            ':input[name="exposed_form_options[bef][filter][' . $filter->field . '][configuration][options_show_only_used]"]' => [
              'checked' => TRUE,
            ],
          ],
        ],
      ],
      'options_show_items_count' => [
        '#type' => 'checkbox',
        '#title' => t('Show option items count'),
        '#description' => t('Show the number of items that will be filtered by each option. Instead of hiding it completely.'),
        '#default_value' => !empty($settings['options_show_items_count']),
        '#states' => [
          'visible' => [
            ':input[name="exposed_form_options[bef][filter][' . $filter->field . '][configuration][options_show_only_used]"]' => [
              'checked' => TRUE
            ],
          ],
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function exposedFormAlter(ViewExecutable &$current_view, FilterPluginBase $filter, array $settings, array &$form, FormStateInterface &$form_state): void {
    // Check that filter is exposed and module functionality enabled.
    if (!$filter->isExposed() || empty($settings['options_show_only_used'])) {
      return;
    }

    $identifier = $filter->options['is_grouped'] ? $filter->options['group_info']['identifier'] : $filter->options['expose']['identifier'];

    // Ensure that request not from this function.
    if (!empty($current_view->selective_filter)) {
      return;
    }

    // Prepare variables for settings.
    $skip_other_filters = empty($settings['options_show_only_used_filtered']);
    $hide_when_empty = !empty($settings['options_hide_when_empty']);
    $show_items_count = !empty($settings['options_show_items_count']);

    /** @var \Drupal\views\ViewExecutable $view */
    $view = Views::getView($current_view->id());
    $view->selective_filter = TRUE;
    $view->setArguments($current_view->args);

    // Support views with contextual filters and exposed form blocks.
    if (
      !empty($current_view->display_handler->getOption('exposed_block'))
      && !empty($current_view->argument)
    ) {
      $args = \Drupal::routeMatch()->getRawParameters()->all();
      if (
        !empty($args['view_id'])
        && !empty($args['display_id'])
        && $args['view_id'] === $current_view->id()
        && $args['display_id'] === $current_view->current_display
      ) {
        $view->setArguments($args);
      }
    }

    $view->setDisplay($current_view->current_display);
    $view->preExecute();

    // Support views with required input.
    if (!empty($view->display_handler->getPlugin('exposed_form')->options['bef']['general']['input_required'])) {
      $view->display_handler->getPlugin('exposed_form')->options['bef']['general']['input_required'] = FALSE;
    }

    // Include all results of a view, ignoring items_per_page that
    // are set in view itself or in one of `views_pre_view` hooks,
    // which are executed in `$view->preExecute()`.
    $view->setItemsPerPage(0);

    // Query parameters can override default results.
    // Save original query and replace with one without parameters.
    $query_param = $view->getRequest()->query;
    $query_param_orig = clone $query_param;

    // Disable per page param.
    $query_param->remove('items_per_page');

    // Skip exposed filters values for full result.
    foreach ($query_param->keys() as $key) {
      // Unset current filter value from input to avoid only one option,
      // in other way current filter value will restrict himself.
      if (!$skip_other_filters && $key !== $identifier) {
        continue;
      }

      $query_param->remove($key);
    }

    // Set modified query
    $view->getRequest()->query = $query_param;

    // Execute modified query.
    $view->execute();

    // Restore parameters for main query.
    $view->getRequest()->query = $query_param_orig;

    // Determinate target element.
    $element = &$form[$identifier];

    // If no results when future processing is haven't any sense.
    if (empty($view->result)) {
      $element['#access'] = $hide_when_empty;
      return;
    }

    $hierarchy = !empty($filter->options['hierarchy']);
    $relationship = $filter->options['relationship'];
    $field_id = $filter->definition['field_name'] ?? NULL;

    if ($filter instanceof Bundle) {
      $field_id = $filter->options['field'];
    }

    \Drupal::moduleHandler()->alter('selective_better_exposed_filters_field_id', $field_id, $view, $filter);

    if (!$field_id) {
      throw new \Exception('Filter field ID could not be determined.');
    }

    $ids = [];
    $relationship_count = [];

    // Avoid illegal choice.
    $user_value = $form_state->getUserInput()[$identifier] ?? NULL;

    if ($user_value !== NULL) {
      if (is_array($user_value)) {
        $ids = $user_value;
      }
      else {
        $ids[$user_value] = $user_value;
      }
    }

    foreach ($view->result as $row) {
      $entity = $row->_entity ?? NULL;

      // Get entity from relation.
      if ($relationship != 'none') {
        $entity = $row->_relationship_entities[$relationship] ?? NULL;
      }

      // Get entity from object.
      if (!$entity) {
        $entity = $row->_object->getEntity();
      }

      // Load entity translation.
      if (
        $entity instanceof TranslatableInterface
        && !empty($row->node_field_data_langcode)
        && $entity->hasTranslation($row->node_field_data_langcode)) {
        $entity = $entity->getTranslation($row->node_field_data_langcode);
      }

      if (
        !($entity instanceof FieldableEntityInterface)
        || !$entity->hasField($field_id)
      ) {
        continue;
      }

      $item_values = $entity->get($field_id)->getValue();

      if (empty($item_values)) {
        continue;
      }

      foreach ($item_values as $item_value) {
        if (array_key_exists('target_id', $item_value)) {
          $id = $item_value['target_id'];
          $relationship_count[$id] = isset($relationship_count[$id]) ? $relationship_count[$id] + 1 : 1;
          $ids[$id] = $id;

          if (!$hierarchy) {
            continue;
          }

          $parents = \Drupal::service('entity_type.manager')
            ->getStorage("taxonomy_term")
            ->loadAllParents($id);

          /** @var \Drupal\taxonomy\TermInterface $term */
          foreach ($parents as $term) {
            $ids[$term->id()] = $term->id();
            $relationship_count[$term->id()] = !empty($relationship_count[$term->id()]) ? $relationship_count[$term->id()] + 1 : 1;
          }
          continue;
        }

        if (array_key_exists('value', $item_value)) {
          $id = $item_value['value'];
          $ids[$id] = $id;
          $relationship_count[$id] = !empty($relationship_count[$id]) ? $relationship_count[$id] + 1 : 1;
        }
      }
    }

    // When no options there is nothing to do.
    if (empty($element['#options'])) {
      $element['#access'] = $hide_when_empty;
      return;
    }

    foreach ($element['#options'] as $key => $option) {
      if ($key === 'All') {
        continue;
      }

      $target_id = $key;
      if (is_object($option) && !empty($option->option)) {
        $target_id = array_keys($option->option);
        $target_id = reset($target_id);
      }

      if (!in_array($target_id, $ids)) {
        unset($element['#options'][$key]);
        continue;
      }

      if (!$show_items_count) {
        continue;
      }

      $count = $relationship_count[$target_id] ?? 0;
      $element['#options'][$key] =  $element['#options'][$key] . ' (' . $count . ')';
    }

    // Make the element size fit with the new number of options.
    if (isset($element['#size'])) {
      if (count($element['#options']) >= 2 && count($element['#options']) < $element['#size']) {
        $element['#size'] = count($element['#options']);
      }
    }

    // Process "Hide when empty" behavior.
    if (count($element['#options']) == 1 && array_key_exists('All', $element['#options'])) {
      $element['#access'] = $hide_when_empty;
    }
  }

}
